// Copyright 2022 schukai GmbH
// SPDX-License-Identifier: AGPL-3.0

package configuration

import (
	"github.com/fsnotify/fsnotify"
	"os"
	"path"
)

func (s *Settings[C]) initWatch() *Settings[C] {

	var err error

	errorCount := len(s.errors)
	defer func() {
		if len(s.errors) > errorCount {
			s.notifyErrorHooks()
		}
	}()

	if s.fileWatch.watcher != nil {
		s.errors = append(s.errors, WatchAlreadyInitializedError)
		return s
	}

	s.fileWatch.watcher, err = fsnotify.NewWatcher()
	if err != nil {
		s.errors = append(s.errors, err)
		return s
	}

	return s

}

func (s *Settings[C]) StopWatching() *Settings[C] {

	s.fileWatch.Lock()
	defer s.fileWatch.Unlock()

	errorCount := len(s.errors)
	defer func() {
		if len(s.errors) > errorCount {
			s.notifyErrorHooks()
		}
	}()

	if s.fileWatch.watcher == nil {
		s.errors = append(s.errors, WatchNotInitializedError)
		return s
	}

	if !s.fileWatch.onWatch {
		s.errors = append(s.errors, WatchNotRunningError)
		return s
	}

	s.fileWatch.cancelWatch <- true

	return s
}

func (s *Settings[C]) buildWatchList() *Settings[C] {

	s.fileWatch.Lock()
	defer s.fileWatch.Unlock()

	s.fileWatch.watchList = make(map[string]bool)

	for _, d := range s.files.directories {
		fn := path.Join(d, s.files.name+s.files.format.Extension())
		s.fileWatch.watchList[fn] = true
	}

	for _, f := range s.files.files {
		s.fileWatch.watchList[f.path] = true
	}

	return s
}

// Watch the given file for changes
func (s *Settings[C]) Watch() *Settings[C] {

	s.buildWatchList()

	s.fileWatch.Lock()
	defer s.fileWatch.Unlock()

	errorCount := len(s.errors)
	defer func() {
		if len(s.errors) > errorCount {
			s.notifyErrorHooks()
		}
	}()

	if s.fileWatch.watcher == nil {
		s.initWatch()
	}

	if s.fileWatch.watchList == nil {
		s.errors = append(s.errors, WatchListNotInitializedError)
		return s
	}

	if s.fileWatch.onWatch == true {
		s.errors = append(s.errors, WatchAlreadyRunningError)
		return s
	}

	s.fileWatch.onWatch = true
	s.fileWatch.cancelWatch = make(chan bool)

	// remove all files from the watch list
	for _, file := range s.fileWatch.watcher.WatchList() {
		s.fileWatch.watcher.Remove(file)
	}

	// add all files to the watch list
	for filePath := range s.fileWatch.watchList {

		fileInfo, err := os.Stat(filePath)
		if err != nil {
			s.errors = append(s.errors, err)
			continue
		}

		if fileInfo.IsDir() {
			err = s.fileWatch.watcher.Add(filePath)
		} else {
			err = s.fileWatch.watcher.Add(path.Dir(filePath))
		}

		if err != nil {
			s.errors = append(s.errors, err)
		}
	}

	if len(s.fileWatch.watcher.WatchList()) == 0 {
		s.errors = append(s.errors, NoFilesToWatchError)
		s.Unlock()
		return s
	}

	go func() {
	finished:

		for {
			select {
			case event, ok := <-s.fileWatch.watcher.Events:
				if !ok {
					return
				}

				_, exist := s.fileWatch.watchList[event.Name]
				if !exist {
					continue
				}

				if event.Op&fsnotify.Write == fsnotify.Write {
					s.Import()
				} else if event.Op&fsnotify.Remove == fsnotify.Remove {
					s.Import()
				} else if event.Op&fsnotify.Rename == fsnotify.Rename {
					s.Import()
				}

			case err, ok := <-s.fileWatch.watcher.Errors:
				if !ok {
					return
				}
				s.errors = append(s.errors, err)

			case <-s.fileWatch.cancelWatch:
				break finished
			}
		}

		s.fileWatch.onWatch = false
	}()

	return s

}
